// Copyright 2009, 2010, 2011 Christopher Smith
// 
// This file is part of smithers.algorithms.
// 
// smithers.algorithms is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.algorithms is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.algorithms. If not, see <http://www.gnu.org/licenses/>.

package smithers.algorithms;

/**
 * Implements a collection of useful matrix related algorithms. The methods of
 * this class all require that the matrix arguments be rectangular; for
 * efficiency they do not check this, hence if the arguments are invalid, they
 * will likely throw {@code ArrayIndexOutOfBoundsException} or {@code
 * NullPointerException}
 */
public class Matrix {
    
    private Matrix() {}
    
    /**
     * Multiplies two matrices together. The first matrix must have as many
     * columns as the second has rows.
     * 
     * @param matrix1 a matrix
     * @param matrix2 another matrix
     * @return {@code matrix1 * matrix2}
     */
    public static float[][] product(float[][] matrix1, float[][] matrix2) {
        final int l = matrix1.length;
        final int m = matrix1[0].length;
        final int n = matrix2[0].length;
        float[][] result = new float[l][n];
        for (int i=0; i<l; i++) for (int j=0; j<n; j++) {
            result[i][j] = 0;
            for (int k=0; k<m; k++)
                result[i][j] += matrix1[i][k] * matrix2[k][j];
        }
        return result;
    }
    
    /**
     * Multiplies a matrix by a column vector. The vector must have as many
     * elements as the matrix has columns.
     * 
     * @param matrix a matrix
     * @param vector a vector
     * @return {@code matrix * vector}
     */
    public static float[] product(float[][] matrix, float[] vector) {
        final int m = matrix.length;
        final int n = matrix[0].length;
        
        float[] result = new float[m];
        for (int i=0; i<m; i++) {
            result[i] = 0;
            for (int j=0; j<n; j++)
                result[i] += matrix[i][j] * vector[j];
        }
        
        return result;
    }
    
    /**
     * Multiplies a row vector by a matrix. The vector must have as many
     * elements as the matrix has rows.
     * 
     * @param vector a vector
     * @param matrix a matrix
     * @return {@code vector * matrix}
     */
    public static float[] product(float[] vector, float[][] matrix) {
        final int m = matrix.length;
        final int n = matrix[0].length;
        
        float[] result = new float[n];
        for (int j=0; j<n; j++) {
            result[j] = 0;
            for (int i=0; i<m; i++)
                result[j] += vector[i] * matrix[i][j];
        }
        
        return result;
    }
    
    /**
     * Row reduce a matrix in-place. Uses Gauss-Jordan elimination with scaled
     * partial pivoting.
     * 
     * @param matrix the matrix to reduce
     */
    public static void rowReduce(float[][] matrix) {
        final int rowCount = matrix.length;
        final int colCount = matrix[0].length;
        int elimRow = 0;
        int i, j;
        
        for (int elimCol=0; elimCol<colCount; elimCol++) {
            // Find the row to use as a pivot.
            float pivotValue = 0;
            int pivotRow = -1;
            for (i=elimRow; i<rowCount; i++) {
                float newPivot = 0;
                for (j=elimCol; j<colCount; j++)
                    newPivot += matrix[i][j] * matrix[i][j];
                newPivot = matrix[i][elimCol] * matrix[i][elimCol] / newPivot;
                if (newPivot > pivotValue) {
                    pivotRow = i;
                    pivotValue = newPivot;
                }
            }
            // Unable to eliminate in this column, try the next column.
            if (pivotRow == -1) continue;
            // Pivot.
            if (pivotRow != elimRow) {
                for (j=elimCol; j<colCount; j++) {
                    float temp = matrix[elimRow][j];
                    matrix[elimRow][j] = matrix[pivotRow][j];
                    matrix[pivotRow][j] = temp;
                }
            }
            pivotValue = matrix[elimRow][elimCol];
            // Reduce this row (ensure leading element is 1).
            for (j=elimCol; j<colCount; j++)
                matrix[elimRow][j] /= pivotValue;
            // Reduce the other rows (element in column elimCol must be 0).
            for (i=0; i<rowCount; i++) {
                if (i == elimRow) continue;
                pivotValue = matrix[i][elimCol];
                for (j=elimCol; j<colCount; j++)
                    matrix[i][j] -= matrix[elimRow][j] * pivotValue;
            }
            
            if (++elimRow == rowCount)
                break;
        }
    }
    
    /**
     * Computes an LQ decomposition of a matrix. The matrix must have at least
     * as many columns as rows. This computes L and Q such that A = LQ, with L
     * lower triangular and Q orthogonal. Uses the modified Gram-Schmidt
     * process, further modified to still work if A is not square or if the
     * rows of A are not linearly independant; in the latter case the matrix L
     * will have one or more zero entries on the diagonal.
     * 
     * @param matrix the matrix to decompose
     * @param l      the matrix in which to store L; must have the same
     *               dimensions as {@code matrix}
     * @param q      the matrix in which to store Q; must be square of
     *               dimension equal to the number of columns of {@code matrix}
     */
    public static void decomposeLQ(float[][] matrix, float[][] l, float[][] q) {
        final int rowCount = matrix.length;
        final int colCount = matrix[0].length;
        int i, j;
        
        for (int row=0; row<rowCount; row++) {
            for (j=0; j<colCount; j++)
                q[row][j] = matrix[row][j];
            orthonormaliseRow(l, q, row);
        }
        
        // Make up any orthonormal vectors for the final rows
        for (int row=rowCount; row<colCount; row++)
            inventRow(q, row);
    }
    
    // Make the given row of q unit and orthogonal to those above it.
    private static void orthonormaliseRow(float[][] l, float[][] q, int row) {
        final int n = q.length;
        int i, j;
        
        // Make this row orthogonal to the rows above.
        for (i=0; i<row; i++) {
            float dot = 0;
            for (j=0; j<n; j++)
                dot += q[row][j] * q[i][j];
            l[row][i] = dot;
            for (j=0; j<n; j++)
                q[row][j] -= dot * q[i][j];
        }
        // Normalise this row.
        float norm = 0;
        for (j=0; j<n; j++)
            norm += q[row][j] * q[row][j];
        norm = (float)Math.sqrt(norm);
        l[row][row] = norm;
        if (norm == 0) {
            // Row is zero - was a linear combination of the rows above.
            // Make up any orthonormal row.
            inventRow(q, row);
        } else {
            for (j=0; j<n; j++)
                q[row][j] /= norm;
        }
        for (i=row+1; i<n; i++)
            l[row][i] = 0;
    }
    
    // Invent a row in q which is unit and orthogonal to those above it.
    private static void inventRow(float[][] q, int row) {
        final int n = q.length;
        int bestCol=-1;
        float bestNorm = 0;
        int i, j;
        
        for (int col=0; col<n; col++) {
            // Try normalising from a 1 in column col
            for (j=0; j<n; j++)
                q[row][j] = 0;
            q[row][col] = 1;
            // Make this row orthongonal to the rows above.
            for (i=0; i<row; i++) {
                float dot = 0;
                for (j=0; j<n; j++)
                    dot += q[row][j] * q[i][j];
                for (j=0; j<n; j++)
                    q[row][j] -= dot * q[i][j];
            }
            // Find the norm
            float norm = 0;
            for (j=0; j<n; j++)
                norm += q[row][j] * q[row][j];
            if (norm > bestNorm) {
                bestCol = col;
                bestNorm = norm;
            }
        }
        
        // Orthonormalise from a 1 in column col
        for (j=0; j<n; j++)
            q[row][j] = 0;
        q[row][bestCol] = 1;
        // Make this row orthogonal to the rows above.
        for (i=0; i<row; i++) {
            float dot = 0;
            for (j=0; j<n; j++)
                dot += q[row][j] * q[i][j];
            for (j=0; j<n; j++)
                q[row][j] -= dot * q[i][j];
        }
        // Normalise this row.
        float norm = 0;
        for (j=0; j<n; j++)
            norm += q[row][j] * q[row][j];
        norm = (float)Math.sqrt(norm);
        for (j=0; j<n; j++)
            q[row][j] /= norm;
    }
    
}
