// Copyright 2009 Christopher Smith
// 
// This file is part of smithers.calculator.
// 
// smithers.calculator is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.calculator is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.calculator. If not, see <http://www.gnu.org/licenses/>.

package smithers.calculator;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.LinkedList;
import java.util.List;

/**
 * A calculator.
 */
public class Calculator {
    
    CalcObject mem = new CalcDouble(0);
    
    /**
     * Parses the given expression.
     * 
     * @param input the expression to evaluate
     * @return the result of the evaluation
     */
    public String parse(String input) {
        return reparse(input).toString();
    }
    
    CalcObject parse(CalcString input) {
        int i;
        final String str = input.value;
        
        // Binary operators
        String[] temp;
        temp = stringSplit(str, '+');
        if (temp.length>1) return reparse("Add", temp);
        temp = stringSplit(str, '-');
        if (temp.length>1) return reparse("Sub", temp);
        temp = stringSplit(str, '*');
        if (temp.length>1) return reparse("Mul", temp);
        temp = stringSplit(str, '/');
        if (temp.length>1) return reparse("Div", temp);
        temp = stringSplit(str, '^');
        if (temp.length>1) return reparse("Pow", temp);
        
        //Unary operators
        if (str.charAt(0) == '+') return reparse(str.substring(1));
        if (str.charAt(0) == '-') return reparse("Neg", str.substring(1));
        
        //Functions
        if (str.matches(".*?\\(.*\\)")) {
            String head = str.substring(0, str.indexOf("("));
            String[] args = stringSplit(str.substring(str.indexOf("(")+1, str.length()-1));
            CalcObject[] argvals = new CalcObject[args.length];
            for (i=0; i<args.length; i++) argvals[i] = reparse(args[i]);
            if (head.equalsIgnoreCase("")) {
                return argvals[0];
            } else if (head.equalsIgnoreCase("Add")) {
                return MethodsArithmetic.add(argvals);
            } else if (head.equalsIgnoreCase("Sub")) {
                return MethodsArithmetic.sub(argvals);
            } else if (head.equalsIgnoreCase("Mul")) {
                return MethodsArithmetic.mul(argvals);
            } else if (head.equalsIgnoreCase("Div")) {
                return MethodsArithmetic.div(argvals);
            } else if (head.equalsIgnoreCase("Pow")) {
                return MethodsArithmetic.pow(argvals);
            } else if (head.equalsIgnoreCase("Sqrt")) {
                return MethodsArithmetic.sqrt(argvals);
            } else if (head.equalsIgnoreCase("Neg")) {
                return MethodsArithmetic.neg(argvals);
            } else if (head.equalsIgnoreCase("Inv")) {
                return MethodsArithmetic.inv(argvals);
                
            } else if (head.equalsIgnoreCase("Exp")) {
                return MethodsAnalytic.exp(argvals);
            } else if (head.equalsIgnoreCase("Log")) {
                return MethodsAnalytic.log(argvals);
            } else if (head.equalsIgnoreCase("Sin")) {
                return MethodsAnalytic.sin(argvals);
            } else if (head.equalsIgnoreCase("Cos")) {
                return MethodsAnalytic.cos(argvals);
            } else if (head.equalsIgnoreCase("Tan")) {
                return MethodsAnalytic.tan(argvals);
            } else {
                throw new RuntimeException("Unknown function: " + head);
            }
        }
        
        //Constants
        if (str.equalsIgnoreCase("e")) return new CalcDouble(Math.E);
        if (str.equalsIgnoreCase("Pi")) return new CalcDouble(Math.PI);
        if (str.equalsIgnoreCase("i")) return new CalcComplexDouble(0,1);
        
        //Literals
        if (str.charAt(str.length()-1)=='i') return new CalcComplexDouble(0, Double.parseDouble(str.substring(0, str.length()-1)));
        return new CalcDouble(Double.parseDouble(str));
    }
    
    CalcObject reparse(String head, String... body) {
        String str = head;
        if (body.length > 0) {
            str += "(";
            for (String temp : body) str += temp + ",";
            str = str.substring(0, str.length()-1) + ")";
        }
        return parse(new CalcString(str));
    }
    
    String[] stringSplit(String str) {
        return stringSplit(str, ',');
    }
    
    String[] stringSplit(String str, char sep) {
        List<String> result = new LinkedList<String>();
        int i=0, j=0, b=0;
        for (j=0; j<str.length(); j++) {
            if (str.charAt(j)=='(') b++;
            if (str.charAt(j)==')') b--;
            if (str.charAt(j)==sep && b==0 && j>0 && ! str.substring(j-1,j).matches("[\\Q+*-/^eE\\E]")) {
                result.add(str.substring(i,j));
                i=j+1;
            }
        }
        result.add(str.substring(i));
        return result.toArray(new String[0]);
    }
    
    /**
     * Runs the calculator program. If arguments are given, they are each
     * evaluated in turn. Otherwise the program reads expressions from standard
     * input until a blank line is read.
     * 
     * @param args the arguments, if any
     */
    public static void main(String[] args) throws Exception {
        Calculator c = new Calculator();
        if (args.length>0) {
            for (String arg : args) {
                if (arg.equals("-h") || arg.equals("--help")) {
                    System.out.println("smithers.calculator.Calculator -- simple calculator utility.");
                    System.out.println("Usage:");
                    System.out.println("        java smithers.calculator.Calculator [options...] [expressions...]");
                    System.out.println("If expressions are given, evaluate them and exit, otherwise evaluate expressions from standard input until a blank line is given.");
                    System.out.println("");
                    System.out.println("Options:");
                    System.out.println("  -h --help");
                    System.out.println("        Print this help and exit.");
                    System.out.println("");
                    System.out.println("Supported syntax and functions to follow...");
                    return;
                }
                System.out.println(c.parse(arg));
            }
        } else {
            String arg = "";
            BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
            arg = in.readLine();
            while (!arg.equals("")) {
                System.out.println(c.parse(arg)+"\n");
                arg = in.readLine();
            }
        }
    }
    
}
