// Copyright 2009, 2010, 2011 Christopher Smith
// 
// This file is part of smithers.java3d.
// 
// smithers.java3d is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.java3d is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.java3d. If not, see <http://www.gnu.org/licenses/>.

package smithers.java3d;

import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelListener;
import java.util.List;
import java.util.ArrayList;

/**
 * Represents a drawable 3D object. Objects of classes extending this can be
 * added to a Panel3D and will be drawn upon it.
 */
public abstract class Drawable3D {
    
    List<MouseListener> mouseListeners = new ArrayList<MouseListener>();
    List<MouseMotionListener> mouseMotionListeners = new ArrayList<MouseMotionListener>();
    List<MouseWheelListener> mouseWheelListeners = new ArrayList<MouseWheelListener>();
    
    /**
     * The 3-d polygon which this object fills up.
     */
    protected Polygon3D poly;
    
    /**
     * Sole constructor.
     */
    public Drawable3D() {}
    
    /**
     * Returns the 3-d polygon which this object fills up.
     * 
     * @return the object's polygon
     */
    public Polygon3D getShape() {
        return poly;
    }
    
    /**
     * Draws this object. This method should draw the object, as transformed by
     * {@code trans}, on {@code g}. The method should return a {@code Shape},
     * indicating the region which it covered. This is to enable the container
     * to deliver mouse events.
     * 
     * @param g     the graphics object on which to draw
     * @param trans the transformation to use
     * @return the region of space which was filled
     */
    public abstract Shape drawOn(Graphics2D g, Transformation trans);
    
    /**
     * Registers the specified listener to receive mouse events from this
     * component.
     * 
     * @param l the listener
     */
    public void addMouseListener(MouseListener l) {
        if (l != null)
            mouseListeners.add(l);
    }
    
    /**
     * Removes the specified listener from receiving mouse events from this
     * component.
     * 
     * @param l the listener
     */
    public void removeMouseListener(MouseListener l) {
        if (l != null)
            mouseListeners.remove(l);
    }
    
    /**
     * Registers the specified listener to receive mouse motion events from
     * this component.
     * 
     * @param l the listener
     */
    public void addMouseMotionListener(MouseMotionListener l) {
        if (l != null)
            mouseMotionListeners.add(l);
    }
    
    /**
     * Removes the specified listener from receiving mouse motion events from
     * this component.
     * 
     * @param l the listener
     */
    public void removeMouseMotionListener(MouseMotionListener l) {
        if (l != null)
            mouseMotionListeners.remove(l);
    }
    
    /**
     * Registers the specified listener to receive mouse wheel events from this
     * component.
     * 
     * @param l the listener
     */
    public void addMouesWheelListener(MouseWheelListener l) {
        if (l != null)
            mouseWheelListeners.add(l);
    }
    
    /**
     * Removes the specified listener from receiving mouse wheel events from
     * this component.
     * 
     * @param l the listener
     */
    public void removeMouesWheelListener(MouseWheelListener l) {
        if (l != null)
            mouseWheelListeners.remove(l);
    }
    
}
