// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;

public class HanjieGraphics implements PuzzleGraphics {
    
    public static final Color COLOUR_UNKNOWN = Color.LIGHT_GRAY;
    public static final Color COLOUR_BORDERS = Color.DARK_GRAY;
    
    private Hanjie puzzle;
    private HanjieState state;
    private boolean hasSolver;
    private HanjieSolver solver;
    private int[] coords = {0, 0, 0, 0, 0, 0};
    
    public HanjieGraphics(HanjieState state) {
        this.state = state;
        puzzle = state.getPuzzle();
        hasSolver = false;
    }
    
    public HanjieGraphics(HanjieSolver solver) {
        this.solver = solver;
        state = solver.getState();
        puzzle = solver.getPuzzle();
        hasSolver = true;
    }
    
    private Color getColour(int index) {
        if (index < 0) return (COLOUR_UNKNOWN);
        if (!puzzle.hasColour) return index > 0 ? Color.BLACK : Color.WHITE;
        return puzzle.colours[index];
    }
    
    public Rectangle draw(Graphics g, Rectangle r) {
        int i, j;
        String str;
        int clueWidth=0, clueHeight=0;
        int maxClue=0;
        int[] clue, colour = null;
        
        for (i=0; i<puzzle.height; i++) {
            clue = puzzle.getRowClue(i);
            if (clue.length > clueWidth) clueWidth = clue.length;
            for (j=0; j<clue.length; j++)
                if (clue[j] > maxClue) maxClue = clue[j];
        }
        for (i=0; i<puzzle.width; i++) {
            clue = puzzle.getColumnClue(i);
            if (clue.length > clueHeight) clueHeight = clue.length;
            for (j=0; j<clue.length; j++)
                if (clue[j] > maxClue) maxClue = clue[j];
        }
        int totHeight = puzzle.height + clueHeight, totWidth = puzzle.width + clueWidth;
        int gridWidth = (r.width - 1) / totWidth;
        int gridHeight = (r.height - 1) / totHeight;
        int gridSize = gridWidth < gridHeight ? gridWidth : gridHeight;
        int boxSize = gridSize - 1;
        int x = r.x + (r.width + gridSize * (clueWidth - puzzle.width) - 1)/2;
        int y = r.y + (r.height + gridSize * (clueHeight - puzzle.height) - 1)/2;
        
        int typeSize, textRelY=0;
        FontMetrics fm = g.getFontMetrics();
        str = Integer.toString(maxClue);
        for (typeSize = 24; typeSize >= 0; typeSize--) {
            g.setFont(new Font("Dialog", Font.BOLD, typeSize));
            fm = g.getFontMetrics();
            int textWidth = fm.stringWidth(str);
            int textAscent = fm.getAscent();
            
            if (textWidth < boxSize && textAscent < boxSize) {
                textRelY = boxSize/2 + textAscent/2;
                break;
            }
        }
        
        g.setColor(COLOUR_BORDERS);
        g.fillRect(x - gridSize * clueWidth, y - gridSize * clueHeight, gridSize * totWidth + 1, gridSize * totHeight + 1);
        g.setColor(COLOUR_UNKNOWN);
        g.fillRect(x - gridSize * clueWidth + 1, y - gridSize * clueHeight + 1, gridSize * clueWidth - 1, gridSize * clueHeight - 1);
        //Draw row clues
        for (i=0; i<puzzle.height; i++) {
            g.setColor(COLOUR_UNKNOWN);
            g.fillRect(x - gridSize * clueWidth + 1, y + i * gridSize + 1, gridSize * clueWidth - 1, boxSize);
            clue = puzzle.getRowClue(i);
            if (puzzle.hasColour) colour = puzzle.getRowColour(i);
            for (j=0; j<clue.length; j++) {
                g.setColor(puzzle.hasColour ? puzzle.colours[colour[j]] : Color.BLACK);
                str = Integer.toString(clue[j]);
                g.drawString(str,
                             x - gridSize*(clue.length-j) + 1 + boxSize/2 - fm.stringWidth(str)/2,
                             y + i*gridSize + 1 + textRelY);
            }
        }
        //Draw column clues
        for (i=0; i<puzzle.width; i++) {
            g.setColor(COLOUR_UNKNOWN);
            g.fillRect(x + i * gridSize + 1, y - gridSize * clueHeight + 1, boxSize, gridSize * clueHeight - 1);
            clue = puzzle.getColumnClue(i);
            if (puzzle.hasColour) colour = puzzle.getColumnColour(i);
            for (j=0; j<clue.length; j++) {
                g.setColor(puzzle.hasColour ? puzzle.colours[colour[j]] : Color.BLACK);
                str = Integer.toString(clue[j]);
                g.drawString(str,
                             x + i * gridSize + 1 + boxSize/2 - fm.stringWidth(str)/2,
                             y - gridSize * (clue.length-j) + 1 + textRelY);
            }
        }
        //Draw main grid
        for (i=0; i<puzzle.height; i++) for (j=0; j<puzzle.width; j++) {
            g.setColor(getColour(state.getState(i, j)));
            g.fillRect(x + j * gridSize + 1, y + i * gridSize + 1, boxSize, boxSize);
        }
        
        if (hasSolver && solver.getLastChange() != -1) {
            int lastChange = solver.getLastChange();
            g.setColor(Color.RED);
            if (solver.getLastWasRow()) {
                for (int d=-1; d<2; d++) {
                    g.drawRect(x - gridSize * clueWidth + d, y + gridSize * lastChange + d, gridSize * clueWidth - 2 * d, gridSize - 2 * d);
                    g.drawRect(x + d, y + gridSize * lastChange + d, gridSize * puzzle.width - 2 * d, gridSize - 2 * d);
                }
            } else {
                for (int d=-1; d<2; d++) {
                    g.drawRect(x + lastChange * gridSize + d, y - gridSize * clueHeight + d, gridSize - 2 * d, gridSize * clueHeight - 2 * d);
                    g.drawRect(x + lastChange * gridSize + d, y + d, gridSize - 2 * d, gridSize * puzzle.height - 2 * d);
                }
            }
        }
        
        coords[0] = x-gridSize*clueWidth;
        coords[1] = x;
        coords[2] = x+gridSize*puzzle.width;
        coords[3] = y-gridSize*clueHeight;
        coords[4] = y;
        coords[5] = y+gridSize*puzzle.height;
        return new Rectangle(x-gridSize*clueWidth, y-gridSize*clueHeight, gridSize*(clueWidth+puzzle.width)+1, gridSize*(clueHeight+puzzle.height)+1);
    }
    
    public void mouseClicked(MouseEvent e) {
        int x = e.getX(), y = e.getY();
        if (x >= coords[1] && x < coords[2] && y >= coords[4] && y < coords[5]) {
            x = (x-coords[1]) * puzzle.getWidth() / (coords[2]-coords[1]);
            y = (y-coords[4]) * puzzle.getHeight() / (coords[5]-coords[4]);
            int s = state.getState(y, x);
            if (e.getButton() == MouseEvent.BUTTON1) {
                if (s == -1) s = 1;
                else if (s == 0) s = -1;
                else if (s == puzzle.getColourCount() - 1) s = 0;
                else s++;
            } else {
                if (s == 1) s = -1;
                else if (s == -1) s = 0;
                else if (s == 0) s = puzzle.getColourCount() - 1;
                else s--;
            }
            state.setState(y, x, s);
            if (hasSolver)
                solver.start();
        } else if (hasSolver) {
            if (e.getButton() == MouseEvent.BUTTON1)
                solver.step();
            else if (e.getButton() == MouseEvent.BUTTON2)
                solver.pass();
            else if (e.getButton() == MouseEvent.BUTTON3)
                solver.solve();
        }
        e.getComponent().repaint();
    }
    
    /* Specified by MouseListener */
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    public void mousePressed(MouseEvent e) {}
    public void mouseReleased(MouseEvent e) {}
    
}
